﻿using System;
using System.Text;

namespace SharpChat {
    public class ChatUser : IEquatable<ChatUser> {
        public const int DEFAULT_SIZE = 30;
        public const int DEFAULT_MINIMUM_DELAY = 10000;
        public const int DEFAULT_RISKY_OFFSET = 5;

        public long UserId { get; }
        public string UserName { get; set; }
        public ChatColour Colour { get; set; }
        public int Rank { get; set; }
        public ChatUserPermissions Permissions { get; set; }
        public string NickName { get; set; }
        public ChatUserStatus Status { get; set; }
        public string StatusText { get; set; }

        public string LegacyName {
            get {
                StringBuilder sb = new();

                if(Status == ChatUserStatus.Away)
                    sb.AppendFormat("&lt;{0}&gt;_", StatusText[..Math.Min(StatusText.Length, 5)].ToUpperInvariant());

                if(string.IsNullOrWhiteSpace(NickName))
                    sb.Append(UserName);
                else {
                    sb.Append('~');
                    sb.Append(NickName);
                }

                return sb.ToString();
            }
        }

        public ChatUser(
            long userId,
            string userName,
            ChatColour colour,
            int rank,
            ChatUserPermissions perms,
            string nickName = null,
            ChatUserStatus status = ChatUserStatus.Online,
            string statusText = null
        ) {
            UserId = userId;
            UserName = userName ?? throw new ArgumentNullException(nameof(userName));
            Colour = colour;
            Rank = rank;
            Permissions = perms;
            NickName = nickName ?? string.Empty;
            Status = status;
            StatusText = statusText ?? string.Empty;
        }

        public bool Can(ChatUserPermissions perm, bool strict = false) {
            ChatUserPermissions perms = Permissions & perm;
            return strict ? perms == perm : perms > 0;
        }

        public string Pack() {
            StringBuilder sb = new();

            sb.Append(UserId);
            sb.Append('\t');
            sb.Append(LegacyName);
            sb.Append('\t');
            sb.Append(Colour);
            sb.Append('\t');
            sb.Append(Rank);
            sb.Append(' ');
            sb.Append(Can(ChatUserPermissions.KickUser) ? '1' : '0');
            sb.Append(' ');
            sb.Append(Can(ChatUserPermissions.ViewLogs) ? '1' : '0');
            sb.Append(' ');
            sb.Append(Can(ChatUserPermissions.SetOwnNickname) ? '1' : '0');
            sb.Append(' ');
            sb.Append(Can(ChatUserPermissions.CreateChannel | ChatUserPermissions.SetChannelPermanent, true) ? '2' : (
                Can(ChatUserPermissions.CreateChannel) ? '1' : '0'
            ));

            return sb.ToString();
        }

        public bool NameEquals(string name) {
            return string.Equals(name, UserName, StringComparison.InvariantCultureIgnoreCase)
                || string.Equals(name, NickName, StringComparison.InvariantCultureIgnoreCase)
                || string.Equals(name, LegacyName, StringComparison.InvariantCultureIgnoreCase);
        }

        public override int GetHashCode() {
            return UserId.GetHashCode();
        }

        public override bool Equals(object obj) {
            return Equals(obj as ChatUser);
        }

        public bool Equals(ChatUser other) {
            return UserId == other?.UserId;
        }

        public static string GetDMChannelName(ChatUser user1, ChatUser user2) {
            return user1.UserId < user2.UserId
                ? $"@{user1.UserId}-{user2.UserId}"
                : $"@{user2.UserId}-{user1.UserId}";
        }
    }
}
